from areas.subroutines.array_subroutine import ArraySubroutine
from areas.subroutines.mixed_subroutine import MixedSubroutine
from areas.subroutines.numeric_subroutine import NumericSubroutine
from areas.subroutines.void_subroutine import VoidSubroutine



def cast_to_output(type, value):
    try:
        if type == 'integer':
            return int(value)
        elif type == 'float':
            return float(value)
        
        return value
    except:
        return value


from areas.parameters.parameter import Parameter


class ArrayParameter(Parameter):
    """Array subroutine parameter"""

    def __init__(self, idx, element_type, is_output):
        super().__init__(idx)
        self.element_type = element_type
        if element_type == 'char':
            self.printf_format = 'c'
        elif element_type == 'chari':
            self.element_type = 'char'
            self.printf_format = 'd'
        elif element_type == 'int':
            self.printf_format = 'd'
        elif element_type == 'long':
            self.printf_format = 'ld'
        else:
            self.printf_format = 'f'
        self.is_output = is_output

    def get_prototype_representation(self):
        return '{}* arg{}'.format(self.element_type, self.idx)

    def get_test_declaration_representation(self):
        return '{} arg{}[] = {{{{ {{}} }}}};'.format(self.element_type, self.idx) if self.is_output else ''

    def get_test_call_representation(self):
        return 'arg{}'.format(self.idx) if self.is_output else '({}[]){{{{ {{}} }}}}'.format(self.element_type)

    def get_literal_representantion(self, value):
        return ','.join(map(str, value)) if (self.element_type != 'char'
                                             and self.element_type != 'chari') else','.join(map(lambda x: str(x) if isinstance(x, int) else "'{}'".format(x), value))

    def get_test_call_output_representation(self):
        var_name = 'arg{}'.format(self.idx)
        array_size = 'sizeof({})/sizeof({}[0])'.format(var_name, var_name)
        return '{}{}{}'.format(
            'printf("[");',
            'for(int i = 0; i < {}; ++i) printf("%{}%s", {}[i], (i < {} - 1 ? "," : "\\0"));'.format(
                array_size, self.printf_format, var_name, array_size),
            'printf("]");')\
            if self.is_output else ''



from ast import literal_eval

from areas.parameters import ArrayParameter as Array
from areas.parameters import NumericParameter as Numeric
from areas.parameters import StringParameter as String
from areas.subroutines.subroutine import Subroutine
from areas.util.type_casting import cast_to_output


class ArraySubroutine(Subroutine):
    """Subroutine that returns one or more arrays"""

    def __init__(self, name, parameters, outputs, architecture="arm"):
        super().__init__(name, parameters, architecture)
        self.outputs = outputs

    def get_nr_outputs(self):
        return len(self.outputs)

    def build_test_call(self):
        return '{} {} {} printf("\\n");'.format(\
            # Declare output variables beforehand, so we have access to them after subroutine call
            ''.join([parameter.get_test_declaration_representation() for parameter in self.parameters]),\
            # Actually make subroutine call
            '{}({});'.format(self.name, ','.join([parameter.get_test_call_representation() for parameter in self.parameters])),\
            # Access previously declared variables to print their final values
            'printf("\\n");'.join(filter(lambda x: x != '', [parameter.get_test_call_output_representation() for parameter in self.parameters])))

    def process_parameters(self, parameters):
        for idx, parameter in enumerate(parameters):
            if parameter == 'string':
                self.parameters.append(String(idx, True if idx >= (
                    len(parameters) - len(self.outputs)) else False))
            elif 'array' in parameter:
                self.parameters.append(Array(idx, parameter.replace('array', '').strip(
                ), True if idx >= (len(parameters) - len(self.outputs)) else False))
            else:  # numeric
                self.parameters.append(Numeric(idx, parameter))

    def compare_outputs(self, expected, real, precision):
        if(len(expected) != len(real)):
            return False
        for out_type, exp, re in zip(self.outputs, expected, real):
            if out_type == 'string':
                return exp == re
            else:  # Array
                arr_type = out_type.replace('array', '').strip()
                re = re.replace('-nan', 'None').replace('+nan',
                                                        'None').replace('nan', 'None')
                re_arr = literal_eval(re)
                if(len(exp) != len(re_arr)):
                    return False
                for exp_el, re_el in zip(exp, re_arr):
                    if (arr_type == 'int' or arr_type == 'long') and exp_el != re_el:
                        return False
                    elif re_el is None:
                        return False
                    elif abs(exp_el-re_el) > precision:
                        return False

        return True

    def convert_outputs(self, real):
        outputs = []

        for type, output in zip(self.outputs, real):
            if type == 'string':
                outputs.append(output)
            else:
                array_output = []

                arr_type = type.replace('array', '').strip()

                output = output.replace(
                    '-nan', 'None').replace('+nan', 'None').replace('nan', 'None')

                re_arr = literal_eval(output)

                for array_element in zip(re_arr):
                    if arr_type == 'int' or arr_type == 'long':
                        array_output.append(
                            cast_to_output('integer', array_element))
                    else:
                        array_output.append(
                            cast_to_output('float', array_element))

        return outputs



import json
from datetime import datetime
from re import IGNORECASE, match

from yaml import YAMLError, safe_load

import areas.util.subroutine_integrity as subroutine_integrity
from areas.exception import ToolFileError
from areas.subroutines import ArraySubroutine as Array
from areas.subroutines import MixedSubroutine as Mixed
from areas.subroutines import NumericSubroutine as Numeric
from areas.subroutines import VoidSubroutine as Void
from areas.tester import Tester


class Test:

    def __init__(self, submission_file, subroutine_file=None, tests_file=None,
                 subroutines=None, test_suite=None,
                 timeout=1.0, float_threshold=1e-6, save_to_file=False):
        self.submission_file = submission_file
        self.subroutine_file = subroutine_file
        self.tests_file = tests_file

        # files always override manual definitions
        self.subroutines = subroutines
        self.test_suite = test_suite

        self.timeout = timeout
        self.float_threshold = float_threshold
        self.save_to_file = save_to_file

        self.__read_files()
        self.__parse_inputs()
        self.__build_test_cases()

    def __parse_inputs(self):
        for key in self.subroutines:
            if key not in self.test_suite:
                raise ToolFileError(
                    f"Subroutine {key} is not in the test suite")
            else:
                subroutine = self.subroutines[key]
                test_suite = self.test_suite[key]

                subroutine_integrity.parse_subroutine(subroutine, test_suite)

    def __load_file(self, file):
        ext = file.split('.')[-1]

        if ext == 'json':
            return json.load(open(file))
        elif ext == 'yaml' or ext == 'yml':
            return safe_load(open(file))
        else:
            raise ToolFileError(
                'File extension not supported: {}'.format(ext))

    def __read_files(self):
        try:
            if self.subroutine_file is not None:
                self.subroutines = self.__load_file(self.subroutine_file)

            if self.tests_file is not None:
                self.test_suite = self.__load_file(self.tests_file)
        except IOError as err:
            raise ToolFileError(
                'Could not open: {}, please specify a valid file'.format(str(err)))
        except YAMLError as err:
            raise ToolFileError(
                'Error parsing YAML files: ({}), please correct syntax'.format(str(err)))

    def __build_subroutine(self, name, definition):
        map_to_truth_value = ['array' in ret or ret ==
                              'string' for ret in definition['return']]

        architecture = definition['architecture'] if 'architecture' in definition else "arm"

        if not len(map_to_truth_value):  # No returns - void subroutine
            return Void(name, definition['params'], architecture)
        elif all(map_to_truth_value):  # Only arrays and/or strings as return values
            return Array(name, definition['params'], definition['return'], architecture)
        elif not any(map_to_truth_value):  # Numeric output
            return Numeric(name, definition['params'], definition['return'][0], architecture)
        else:  # Mixed return
            return Mixed(name, definition['params'], definition['return'][0], definition['return'][1:], architecture)

    def __build_test_cases(self):
        self.subroutine_objects = {name: self.__build_subroutine(
            name, definition) for name, definition in self.subroutines.items()}

    def run(self):
        folder_prefix = str(datetime.now().microsecond)

        grading_folder = f'tmp_{folder_prefix}_grading'
        feedback_folder = f'tmp_{folder_prefix}_feedback'

        try:
            tester = Tester(self.subroutine_objects, self.test_suite, grading_folder=grading_folder,
                            feedback_folder=feedback_folder, float_threshold=self.float_threshold,
                            timeout=self.timeout, save_to_file=self.save_to_file)

            submission = tester.grade_submission(self.submission_file)

            filename = match(
                r'(?:.+\/)*(.*)?.*\.zip', self.submission_file, flags=IGNORECASE).group(1)

            return {
                'submission_name': filename,
                'subroutines': submission,
            }
        except Exception as err:
            raise err
        finally:
            import shutil
            shutil.rmtree(grading_folder, ignore_errors=True)
            shutil.rmtree(feedback_folder, ignore_errors=True)



from areas.parameters.parameter import Parameter


class NumericParameter(Parameter):
    """Numeric subroutine parameter"""

    def __init__(self, idx, num_type):
        super().__init__(idx)
        self.num_type = num_type

    def get_prototype_representation(self):
        return '{} arg{}'.format(self.num_type, self.idx)

    def get_test_declaration_representation(self):
        # Not applicable to numeric parameters (they're always only input)
        return ''

    def get_test_call_representation(self):
        return '{}'

    def get_literal_representantion(self, value):
        return str(value)

    def get_test_call_output_representation(self):
        # Not applicable to numeric parameters (when they are output they can be incorporated with the subroutine call)
        return ''



from ast import literal_eval

from areas.parameters import ArrayParameter as Array
from areas.parameters import NumericParameter as Numeric
from areas.parameters import StringParameter as String
from areas.subroutines.subroutine import Subroutine
from areas.util.type_casting import cast_to_output


class NumericSubroutine(Subroutine):
    """Subroutine that returns a single numeric value (e.g., int, long, float, double)"""

    def __init__(self, name, parameters, return_type, architecture="arm"):
        super().__init__(name, parameters, architecture)
        self.c_function_return = return_type

        if return_type == 'int':
            self.printf_format = 'd'
        elif return_type == 'long':
            self.printf_format = 'ld'
        else:
            self.printf_format = 'f'

    def get_nr_outputs(self):
        return 0

    def build_test_call(self):
        return 'printf("%{}\\n", {}({}));'.format(self.printf_format, self.name,
                                                  ','.join([parameter.get_test_call_representation() for parameter in self.parameters]))

    def process_parameters(self, parameters):
        for idx, parameter in enumerate(parameters):
            if parameter == 'string':
                self.parameters.append(String(idx, False))
            elif 'array' in parameter:
                self.parameters.append(
                    Array(idx, parameter.replace('array', '').strip(), False))
            else:  # numeric
                self.parameters.append(Numeric(idx, parameter))

    def compare_outputs(self, expected, real, precision):
        if(len(real) != len(expected)):
            return False
        if self.c_function_return == 'int' or self.c_function_return == 'long':
            return expected[0] == int(real[0])
        else:
            return abs(expected[0] - float(real[0])) <= precision

    def convert_outputs(self, real):
        if self.c_function_return == 'int' or self.c_function_return == 'long':
            return [cast_to_output('integer', real[0])]

        return [cast_to_output('float', real[0])]



class Subroutine:
    """Abstract parent class that defines a generic subroutine and all the methods it should implement"""

    def __init__(self, name, parameters, architecture="arm"):
        self.name = name
        self.parameters_raw = parameters
        # By default, no return value in C function. To be overriden by subclasses when necessary
        self.c_function_return = "void"
        self.parameters = []
        self.architecture = architecture

    def get_nr_outputs(self):
        """Method that returns number of non int outputs"""
        pass

    def build_c_file(self):
        self.process_parameters(self.parameters_raw)
        return '{} {} {}'.format(
            self.build_headers(),
            self.build_prototype(),
            'int main() {{{{ {} return 0;}}}}'.format(self.build_test_call()))

    def build_headers(self):
        return '#include <stdio.h>\n'

    def build_prototype(self):
        return 'extern {} {}({});\n'.format(self.c_function_return, self.name, ','.join(map(lambda p: p.get_prototype_representation(), self.parameters)))

    def build_test_call(self):
        """Method where subroutines implement the calls to test the input data and print out the calls' results"""
        pass

    def process_parameters(self, parameters):
        """Method where subroutines create argument objects for each type. Depending on subroutine type might be an output too."""
        pass

    def compare_outputs(self, expected, real, precision):
        """Method that, given an instance of test outputs, compares to see if the real results match the expected ones"""
        pass

    def convert_outputs(self, real):
        """Method that, given an instance of test outputs, converts it to its real types"""
        pass



ARCHITECTURES = {
    'arm': {
        'compiler': 'aarch64-linux-gnu-gcc',
        'emulator': 'qemu-aarch64-static',
    },
    'riscv': {
        'compiler': 'riscv64-linux-gnu-gcc',
        'emulator': 'qemu-riscv64-static',
    }
}



from areas.exception import ToolFileError
from areas.config.architectures import ARCHITECTURES

AVAILABLE_TYPES = [
    "int",
    "long",
    "float",
    "double",
    "char",
    "chari",
    # arrays
    "char*",
    "string",
    "array int",
    "array long",
    "array float",
    "array double",
    "array char",
    "array chari"
]

PYTHON_COMPATIBLE_TYPES = {
    "int": int,
    "long": int,
    "float": float,
    "double": float,
    "char": str,
    "chari": int,
    # arrays
    "char*": str,
    "string": str,
    "array int": list,
    "array long": list,
    "array float": list,
    "array double": list,
    "array char": list,
    "array chari": list
}


def __parse_subroutine_definition(subroutine):
    if "params" not in subroutine:
        raise ToolFileError("No parameters in subroutine definition")

    if "architecture" in subroutine:
        architecture = subroutine["architecture"].lower()

        if architecture not in ARCHITECTURES:
            raise ToolFileError(
                f"Architecture {architecture} is not supported")

    params = subroutine["params"]
    returns = subroutine["return"] or []

    for param in params:
        if param not in AVAILABLE_TYPES:
            raise ToolFileError(f"Parameter type {param} not supported")

    for return_type in returns:
        if return_type not in AVAILABLE_TYPES:
            raise ToolFileError(f"Return type {return_type} not supported")

    return params, returns


def __parse_test(inputs, outputs, params, returns):
    if len(inputs) != len(params):
        raise ToolFileError(
            "Number of inputs does not match number of parameters")

    if len(outputs) != len(returns):
        raise ToolFileError(
            "Number of outputs does not match number of returns")

    for param, input in zip(params, inputs):
        expected_type = PYTHON_COMPATIBLE_TYPES[param]

        if not isinstance(input, expected_type):
            raise ToolFileError(
                f"Input {input} is not of type {expected_type}")

    for return_type, output in zip(returns, outputs):
        expected_type = PYTHON_COMPATIBLE_TYPES[return_type]

        if not isinstance(output, expected_type):
            raise ToolFileError(
                f"Output {output} is not of type {expected_type}")


def __parse_subroutine_tests(tests, params, returns):
    for test in tests:
        if "inputs" not in test:
            raise ToolFileError("No inputs in test suite")

        if "outputs" not in test:
            raise ToolFileError("No outputs in test suite")

        inputs = test["inputs"]
        outputs = test["outputs"]

        __parse_test(inputs, outputs, params, returns)


def parse_subroutine(subroutine, tests):
    params, returns = __parse_subroutine_definition(subroutine)

    __parse_subroutine_tests(tests, params, returns)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['areas',
 'areas.config',
 'areas.exception',
 'areas.parameters',
 'areas.subroutines',
 'areas.util']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'Unidecode>=1.3.4,<2.0.0', 'pytest>=7.1.3,<8.0.0']

entry_points = \
{'console_scripts': ['test = scripts:test']}

setup_kwargs = {
    'name': 'areas',
    'version': '0.1.1',
    'description': 'ARM64 and RISC-V (extensible) assessment system',
    'long_description': '# areas<!-- omit in toc -->\n\n![GitHub Workflow Status](https://img.shields.io/github/workflow/status/luist18/areas/Deploy%20Docker%20image?logo=github)\n![Docker Image Size (latest by date)](https://img.shields.io/docker/image-size/luist188/areas?logo=docker)\n[![GitHub license](https://img.shields.io/github/license/luist18/areas?color=blue)](https://github.com/luist18/areas/blob/main/LICENSE)\n\n**A**RM64 and **R**ISC-V (**e**xtensible) **A**ssessment **S**ystem.\n\n*areas* is originally a fork from [João Damas\'](https://github.com/cyrilico) [Automatic Observation and (grade) Calculation for (subroutine) Operations tool](https://github.com/cyrilico/aoco-code-correction). It is a tool to automate student\'s grading in the assignments done during the Microprocessor and Personal Computers course unit.\n\n## Differences with the original tool<!-- omit in toc -->\n\nTo ease the communication between the backend server and the tool the output demanded changes. Output `.txt` and `.csv` files are now combined in a more complete `.json` file. Structure of the `.zip` input file is simplified. Unsupported data types such as long and double are now supported. A new input parameter - weight - is introduced.\n\n---\n\n## Table of contents<!-- omit in toc -->\n\n- [1. Installation](#1-installation)\n- [2. Developing](#2-developing)\n- [3. Running](#3-running)\n- [4. Usage](#4-usage)\n- [5. File syntax and structure](#5-file-syntax-and-structure)\n  - [5.1. Available data types](#51-available-data-types)\n    - [5.1.1. Primitive data types](#511-primitive-data-types)\n    - [5.1.2. Array data types](#512-array-data-types)\n  - [5.2. subroutines.yaml](#52-subroutinesyaml)\n  - [5.3. tests.yaml](#53-testsyaml)\n  - [5.4. submission.zip](#54-submissionzip)\n- [6. Results](#6-results)\n\n## 1. Installation\n\nUsing Docker:\n\n```bash\ndocker pull luist188/areas\n```\n\n## 2. Developing\n\nTo develop the tool you must setup a Docker development environment to ease the dependencies installation and setup an isolated environment.\n\n1. Build the Docker development image:\n\n   ```bash\n   docker build -f Dockerfile.dev -t areas .\n   ```\n\n2. Run the image with the shared folder:\n\n   ```bash\n   docker run -it -v $(pwd):/usr/app areas\n   ```\n\n## 3. Running\n\n1. Place the input files inside any directory.\n2. Run the image with a shared volume pointing to the input directory: `docker run -v input:destination -it luist188/areas` (you can learn more about `docker run` usage [here](https://docs.docker.com/engine/reference/run/))\n3. Run the alias command (assure you are using `/bin/bash`) `areas` or run `python main.py` in the tool\'s source.\n\n## 4. Usage\n\n```console\n$ areas [-h] -sr SR -t T -sm SM [SM ...] [-gfd GFD] [-ffd FFD] [-grf GRF] [-tout TOUT] [-fpre FPRE]\n\n$ areas [args]\n\nOptions:\n  --help, -h                Show help                                         [boolean]\n  -sr <subroutines.yaml>    .yaml file containing subroutine declaration      [required] [string]\n  -t <tests.yaml>           .yaml file containing the test cases              [required] [string]\n  -sm <submission.zip...>   .zip files containing user submission             [required] [string array]\n  -gfd <directory>          path to the directory to store temporary files\n    (e.g., compiled binaries)                                                 [default:grading] [string]\n  -ffd <directory>          path to the directory to store the grading for\n    each submission                                                           [default:feedback] [string]\n  -tout <timeout>           float timeout value                               [default:2.0] [float]\n  -fpre <precision>         floating point threshold for comparing floating\n    points in test cases                                                      [default:1e-6] [float]\n```\n\n## 5. File syntax and structure\n\n### 5.1. Available data types\n\n#### 5.1.1. Primitive data types\n\n- `int`\n- `long`\n- `float`\n- `double`\n- `char`\n- `chari` (char represented as an unsgined intenger - similar to char but has to be used when printed characters are not ASCII characters)\n\n#### 5.1.2. Array data types\n\n- `char*/string`\n- `array int`\n- `array long`\n- `array float`\n- `array double`\n- `array char`\n- `array chari`\n\n### 5.2. subroutines.yaml\n\nThe input file for the subroutine declaration has to follow a specific structure and syntax described as follows:\n\n```yaml\nfoo: \n  params: \n    - int\n    - array char\n    - array int\n    - array int\n  return: \n    - int\n    - array int\n\nbar: \n  params: \n    - long\n  return: \n    - long\n```\n\nEach subroutine has an optional parameter to define the subroutine architecture, the syntax is as follows:\n\n```yaml\nfoo: \n  architecture: arm\n  params: \n    - int\n    - array char\n    - array int\n    - array int\n  return: \n    - int\n    - array int\n```\n\nBy default, if the architecture parameter is omitted, the system will assume ARM64 as the subroutine architecture. The available architectures are the following:\n\n- `arm` - ARM64 architecture\n- `riscv` - RISC-V architecture\n\nThe subroutine name has to match the `.s` to test and is case insensitive. Thus, the subroutine `foo` or `bar` is going to check any `.s` file that matches its name case insensitive. All subroutines must contain an array of parameters, `params`, and an array of returns, `return`.\n\n### 5.3. tests.yaml\n\nThe input file for the test cases declaration has to follow a specific structure and syntax described as follows:\n\n```yaml\nbar:\n  - inputs:\n    - 6\n    outputs: \n    - 36\n    weight: 0.5\n  - inputs:\n    - 5\n    outputs: \n    - 25\n    weight: 0.5\n```\n\nThe root declaration of a test case must match the name declared in the `subroutines.yaml` file. Test cases have an array of inputs that has a list of outputs and a test weight. The sum of the test weights must be 1.0.\n\n### 5.4. submission.zip\n\nThe submission `zip` file must contain a `.s` file in its root. For example, for the subroutine `foo` and `bar` the `zip` structure should be as follows:\n\n```tree\nsubmission.zip\n├── foo.s\n└── bar.s\n```\n\n## 6. Results\n\nFor each submission file a `.json` file is created in the feedback directory with the same name of the `.zip` file. The file contains all information about compilation status and test cases. In addition, a simplified version of the result of all submissions is created in a `result.json`. The content of the files look as follows:\n\nFile **submission.json**\n\n```json\n[\n    {\n        "name": "foo",\n        "compiled": true,\n        "ok": true,\n        "passed_count": 2,\n        "test_count": 2,\n        "score": 1,\n        "tests": [\n            {\n                "weight": 1,\n                "run": true,\n                "input": [\n                    6,\n                    ["-", "+", "+", "-", "-", "+"],\n                    [1, 2, 3, 0, 1, -25],\n                    [13, 2, 8, 4, 5, 25]\n                ],\n                "output": [\n                    "0",\n                    ["12", "4", "11", "4", "4", "0"]\n                ],\n                "passed": true\n            }\n        ]\n    },\n    {\n        "name": "bar",\n        "compiled": true,\n        "ok": true,\n        "passed_count": 2,\n        "test_count": 2,\n        "score": 1,\n        "tests": [\n            {\n                "weight": 0.5,\n                "run": true,\n                "input": [\n                    6\n                ],\n                "output": [\n                    "36"\n                ],\n                "passed": true\n            },\n            {\n                "weight": 0.5,\n                "run": true,\n                "input": [\n                    5\n                ],\n                "output": [\n                    "25"\n                ],\n                "passed": true\n            }\n        ]\n    }\n]\n```\n\nFile **result.json**\n\n```json\n[\n    {\n        "submission_name": "submission",\n        "subroutines": [\n            {\n                "name": "foo",\n                "score": 0\n            },\n            {\n                "name": "bar",\n                "score": 0.5\n            }\n        ]\n    },\n    {\n        "submission_name": "submission2",\n        "subroutines": [\n            {\n                "name": "foo",\n                "score": 1\n            },\n            {\n                "name": "bar",\n                "score": 1\n            }\n        ]\n    }\n]\n```\n\n## License<!-- omit in toc -->\n\n[MIT](https://choosealicense.com/licenses/mit/)\n',
    'author': 'Luis Tavares',
    'author_email': 'luistavares10@outlook.pt',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/luist18/areasear',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from ast import literal_eval

from areas.parameters import ArrayParameter as Array
from areas.parameters import NumericParameter as Numeric
from areas.parameters import StringParameter as String
from areas.subroutines.subroutine import Subroutine


class VoidSubroutine(Subroutine):
    """Subroutine that does not return anything, directly or indirectly (i.e., all output is printed directly in assembly)"""

    def __init__(self, name, parameters, architecture="arm"):
        super().__init__(name, parameters, architecture)

    def get_nr_outputs(self):
        return 0

    def build_test_call(self):
        return '{}({}); printf("\\n");'.format(self.name, ','.join([parameter.get_test_call_representation() for parameter in self.parameters]))

    def process_parameters(self, parameters):
        for idx, parameter in enumerate(parameters):
            if parameter == 'string':
                self.parameters.append(String(idx, False))
            elif 'array' in parameter:
                self.parameters.append(
                    Array(idx, parameter.replace('array', '').strip(), False))
            else:  # numeric
                self.parameters.append(Numeric(idx, parameter))

    def compare_outputs(self, expected, real, _):
        truth_values = [exp == re for exp, re in zip(expected, real)]
        return all(truth_values)



import json
from argparse import ArgumentParser
from re import IGNORECASE, match
from sys import exit

from yaml import YAMLError, safe_load

from areas.subroutines import ArraySubroutine as Array
from areas.subroutines import MixedSubroutine as Mixed
from areas.subroutines import NumericSubroutine as Numeric
from areas.subroutines import VoidSubroutine as Void
from areas.tester import Tester as Evaluator


def parse_args():
    """Uses argparse module to create a pretty CLI interface that has the -h by default and that helps the user understand the arguments and their usage
    Returns a dict of "argname":"value"
    """
    parser = ArgumentParser(
        prog="code-correction",
        description="AOCO - Automatic Observation and Correction of (subroutine) Operations")

    parser.add_argument(
        '-sr', help='YAML file containing definition of subroutines', required=True)
    parser.add_argument(
        '-t', help='YAML file containing test cases for evaluation', required=True)
    parser.add_argument(
        '-sm', help='Whitespace separated list of .zip files corresponding to students\' submissions', required=True, nargs='+')
    parser.add_argument(
        '-gfd', help='Folder to store temporary files necessary for grading a submission. Defaults to \'grading\'', default='grading')
    parser.add_argument(
        '-ffd', help='Folder to store submission feedback for each student. Defaults to \'feedback\'', default='feedback')
    parser.add_argument(
        '-grf', help='CSV file to store student grades. Defaults to \'grades.csv\'', default='grades.csv')
    parser.add_argument(
        '-tout', help='Timeout value for when running compiled submissions. Defaults to 2 (seconds)', type=float, default=2)
    parser.add_argument(
        '-fpre', help='Floating point threshold, for when comparing floating point number outputs (accepts values within [VALUE-threshold,VALUE+threshold]). Defaults to 1e-6', type=float, default=1e-6)

    return vars(parser.parse_args())


def build_subroutine(name, definition):
    """Creates a C file template that will run all the test inputs for a given subroutine later
    """
    map_to_truth_value = ['array' in ret or ret ==
                          'string' for ret in definition['return']]
    if not len(map_to_truth_value):  # No returns - void subroutine
        return Void(name, definition['params'])
    elif all(map_to_truth_value):  # Only arrays and/or strings as return values
        return Array(name, definition['params'], definition['return'])
    elif not any(map_to_truth_value):  # Numeric output
        return Numeric(name, definition['params'], definition['return'][0])
    else:  # Mixed return
        return Mixed(name, definition['params'], definition['return'][0], definition['return'][1:])


if __name__ == "__main__":
    args = parse_args()
    try:
        subroutines = safe_load(open(args['sr']))
        test_suite = safe_load(open(args['t']))
    except IOError as err:
        print('Could not open: {}, please specify a valid file'.format(str(err)))
        exit(-1)
    except YAMLError as err:
        print('Error parsing YAML files: ({}), please correct syntax'.format(str(err)))

    subroutine_objs = {name: build_subroutine(
        name, definition) for name, definition in subroutines.items()}

    program_args = {key: args[key]
                    for key in ['gfd', 'ffd', 'grf', 'tout', 'fpre']}
    evaluator = Evaluator(subroutine_objs, test_suite,
                          cmd_line_args=program_args)

    result = []

    for student_submission in args['sm']:
        submission = evaluator.grade_submission(student_submission)

        filename = match(
            r'(?:.+\/)*(.*)?.*\.zip', student_submission, flags=IGNORECASE).group(1)

        subroutines_result = [{
            'name': subroutine['name'],
            'score': subroutine['score']
        } for subroutine in submission]

        result.append({
            'submission_name': filename,
            'subroutines': subroutines_result
        })

    file = open('{}/result.json'.format(program_args['ffd']), 'w')
    json.dump(result, file, indent=4)



import json
import os
import subprocess
from decimal import Decimal
from re import IGNORECASE, match, search
from shutil import rmtree as delete_dir
from zipfile import ZipFile as unzip

import unidecode

from areas.config.architectures import ARCHITECTURES
from areas.exception import *


class Tester:
    """Evaluator class that grades student's submissions based on test input/output comparison and optional code scoring"""

    def __init__(self, subroutines, test_suite, cmd_line_args=None, grading_folder=None,
                 feedback_folder=None, float_threshold=None, timeout=None, save_to_file=True):
        self.temp_grading_folder = grading_folder or cmd_line_args['gfd']
        self.feedback_folder = feedback_folder or cmd_line_args['ffd']
        self.subroutines = subroutines
        self.float_threshold = float_threshold or cmd_line_args['fpre']
        self.timeout = timeout or cmd_line_args['tout']
        self.save_to_file = save_to_file

        self.template_files = [sr.build_c_file()
                               for sr in self.subroutines.values()]

        self.test_outputs = [list(map(lambda test: test['outputs'], test_suite[name]))
                             for name in self.subroutines.keys()]
        self.test_inputs = [list(map(lambda test: test['inputs'], test_suite[name]))
                            for name in self.subroutines.keys()]
        self.test_weights = [list(map(lambda test: test['weight'], test_suite[name]))
                             for name in self.subroutines.keys()]

        if os.path.exists(self.feedback_folder):
            delete_dir(self.feedback_folder)

        if self.save_to_file:
            os.mkdir(self.feedback_folder)

    def extract_submission_files(self, student_submission):
        """Extracts all relevant subroutine files from a submission in order to be graded"""
        with unzip(student_submission, 'r') as zip_file:
            # Get a list of all subroutine file names from the zip
            files = filter(lambda x: x is not None, map(lambda y: match(
                r'(?:\w+/)*([\w\-]+\.s)', y), zip_file.namelist()))
            for file in files:
                with open('{}/{}'.format(self.temp_grading_folder, file.group(1).lower()), mode='wb') as f:
                    f.write(zip_file.read(file.group(0)))

    def get_extra_assembly_to_include(self, subroutine):
        """Obtains a list of subroutines that are used in the given one so that they can be correctly included at compilation time"""
        try:
            with open('{}/{}.s'.format(self.temp_grading_folder, subroutine.lower()), mode='rb') as f:
                calls = list(filter(lambda x: x is not None, map(lambda y: search(
                    rb'bl\s+([\w\-\_]+)', y.strip(), flags=IGNORECASE), f.readlines())))

                calls = set(map(lambda x: x.group(
                    1).lower().decode('utf-8'), calls))

                return [sr for sr in calls if os.path.exists('{}/{}.s'.format(self.temp_grading_folder, sr))]
        except FileNotFoundError:
            return []

    def cleanup_files(self, files):
        for path in files:
            file = open(path, 'r+', encoding='ISO-8859-1')
            new_content = file.read().replace('system', 'blacklisted_word')
            file.seek(0)
            file.write(new_content)
            file.truncate()
            file.close()

    def compile_and_run(self, subroutine, template_file, inp, output_file, architecture="arm"):
        """Compiles and runs, with the generated C template files, the student's submission files.
        Returns a boolean tuple with information of a successful (or not) compilation and a successful (or not) test run"""

        compiler = ARCHITECTURES[architecture]['compiler']
        emulator = ARCHITECTURES[architecture]['emulator']

        sr = self.subroutines[subroutine]
        sr_non_int_outputs = sr.get_nr_outputs()

        # Rearrange parameter-input pair order for format due to possible parameters that are output and, therefore, declared first
        sorted_param_list = list(zip(sr.parameters, inp))
        sorted_param_list = sorted_param_list[-sr_non_int_outputs:] + \
            sorted_param_list[:-sr_non_int_outputs]

        # Build C file from template
        open('{}.c'.format(output_file), 'w').write(
            template_file.format(*[param.get_literal_representantion(inp_literal) for param, inp_literal in sorted_param_list]))

        compilation_files = list(
            map(lambda x: "{}/{}.s".format(self.temp_grading_folder, x.lower()),
                [subroutine, *self.get_extra_assembly_to_include(subroutine)])
        )

        try:
            self.cleanup_files(compilation_files)  # temporary fix
        except Exception as e:
            pass

        # Compile student code alongside generated C file
        compilation_process = subprocess.Popen('{} -o {} {}.c {} -static'.format(
            compiler,
            output_file,
            output_file,
            ' '.join(compilation_files)
        ).split(' '), stderr=subprocess.PIPE)
        compilation_process.wait()
        (_, stderr) = compilation_process.communicate()

        # Throw compilation error
        if compilation_process.returncode != 0:  # If it doesn't even compile, not worth checking any further
            raise CompileError(stderr)

        # Execute and redirect output to temporary .txt file
        real_output_file = '{}.txt'.format(output_file)
        execution_process = subprocess.Popen('timeout {} {} {}'.format(self.timeout, emulator, output_file), stdout=open(
            real_output_file, mode='w', encoding='utf-8'), stderr=subprocess.PIPE, shell=True)
        execution_process.wait()
        (_, stderr) = execution_process.communicate()

        if execution_process.returncode != 0:
            if execution_process.returncode == 124:  # Return code when timeout had to kill the process
                raise RuntimeError(
                    'Failed to run: process took longer than {}s limit'.format(self.timeout))
            else:  # Other reason, but not running properly
                first_index = str(stderr).find('"') + 1
                last_index = str(stderr).rfind('"')

                message = str(stderr)[
                    first_index:last_index].strip()

                raise RuntimeError(
                    'Failed to run: {}\n'.format(message))

    def compare_test_call_output(self, subroutine, output_file, expected_outputs, given_inputs):
        """Compares expected to real outputs, calculating student's score on it"""
        # Read real outputs
        real_outputs = list(map(lambda x: x.decode(
            'utf-8', 'backslashreplace').strip(), open('{}.txt'.format(output_file), 'rb').readlines()))

        # Should return outputs and result (expected if necessary)

        converted_output = self.subroutines[subroutine].convert_outputs(
            real_outputs)

        # Actual comparison
        if not self.subroutines[subroutine].compare_outputs(expected_outputs, real_outputs, self.float_threshold):
            return (False, {
                'output': list(map(lambda x: repr(x).replace("'", ""), real_outputs)),
                'expected': expected_outputs,
                'passed': False
            })

        return (True, {
            'output': list(map(lambda x: repr(x).replace("'", ""), real_outputs)),
            'passed': True
        })

    def grade_submission(self, student_submission):
        """Grades a student's submission by comparing each real output to expected output, writing feedback in failing cases if applicable and final grade to grades csv"""
        # Reset temporary grading folder if for some reason it already exists
        if os.path.exists(self.temp_grading_folder):
            delete_dir(self.temp_grading_folder)
        os.mkdir(self.temp_grading_folder)

        # Extract all assembly files to grading directory
        self.extract_submission_files(student_submission)

        filename = match(
            r'(?:.+\/)*(.*)?.*\.zip', student_submission, flags=IGNORECASE).group(1)

        if self.save_to_file:
            feedback_file = open(
                '{}/{}.json'.format(self.feedback_folder, filename), 'w')

        subroutine_list = []

        for subroutine, template_file, inputs, outputs, weights in zip(self.subroutines.keys(), self.template_files, self.test_inputs, self.test_outputs, self.test_weights):
            # set default compiler as arm
            architecture = self.subroutines[subroutine].architecture or "arm"

            # Subroutine output file template
            output_file = '{}/{}'.format(self.temp_grading_folder,
                                         subroutine.lower())

            passed = 0

            # Creating subroutine object
            subroutine_object = {
                'name': subroutine
            }

            compiled = True
            compile_output = None

            all_tests_passed = True

            tests_list = []

            score = 0

            for inp, out, weight in zip(inputs, outputs, weights):
                # Try to compile and run submission for a specific input case
                # If it fails, grade with zero and move to next subroutine

                test_object = {}

                test_object['weight'] = weight

                try:
                    self.compile_and_run(
                        subroutine, template_file, inp, output_file, architecture)
                except CompileError as compile_error:
                    compile_output = str(compile_error)
                    compiled = False
                    break

                except RuntimeError as runtime_error:
                    all_tests_passed = False

                    test_object['run'] = False
                    test_object['error'] = str(runtime_error)
                    test_object['input'] = inp

                    tests_list.append(test_object)
                    continue

                test_object['run'] = True

                (test_passed, test_compare_object) = self.compare_test_call_output(
                    subroutine, output_file, out, inp)

                if test_passed:
                    passed = passed + 1

                    score = score + int(Decimal(weight) * 1000000)
                else:
                    all_tests_passed = False

                test_object['input'] = inp
                test_object.update(test_compare_object)
                tests_list.append(test_object)

            subroutine_object['compiled'] = compiled

            if not compiled:
                subroutine_object['compile_output'] = compile_output

            subroutine_object['ok'] = compiled and all_tests_passed
            subroutine_object['passed_count'] = passed
            subroutine_object['test_count'] = len(outputs)
            subroutine_object['score'] = float(Decimal(score) / 1000000)
            subroutine_object['tests'] = tests_list

            subroutine_list.append(subroutine_object)

        if self.save_to_file:
            json.dump({'submission_name': filename,
                       'subroutines': subroutine_list}, feedback_file, indent=4)
        # Remove temporary auxiliary folder and write final scores
        # delete_dir(self.temp_grading_folder)

        return subroutine_list
